#!/usr/bin/perl -w
# 
# Reconcile_stock_mysql.pl
#
# author Al Kingsley <al@bananahead.com>
# license	GPL version 2 or later.
# copyright Al Kingsley 2004
# 
#
#	Reconcile stock against transactions
#
#  In General:
#     Pass 1 is sales (decrement stock)
#     Pass 2 is returns (increment stock - if returned to stock)
#     Pass 3 goods in (increment stock)
#     Pass 3 is transfer in / out (inc/dec)
#     Pass 4 is dispatch orders (dec)
#
#  And in keep with a transactional type operation:
#     A synch table entry is made on the start of the operation.
#     The reconcile period is taken from the last closed synch table entry.
#     Reconcile ends with closing of the synch table entry.
#
#  Doesn't do (and shouldn't):
#     Scrapped goods reconcilliation
#     Stock discrepancies
#     Stock take discrepancies
#
#  Limitations:
#     Only handles 1:1 - inventory:product relations.
#
#  To Do:
#     Reconcile products made up from multiple inventory items to fix above.
#

use DBI;
use Getopt::Long;
use POSIX qw(strftime);
use Time::Local;
use Cwd;
use MIME::Lite;
# COMMENT out the use Net::SMTP line you're not using smtp (M$ option)
#use Net::SMTP;

# some options
$email_notification = 'no';
$mailprog           = '/usr/sbin/sendmail -t -oi';
#$mailprog          = '/usr/lib/sendmail -t -oi';
#$mailprog          = "/var/qmail/bin/qmail-inject -h";
$smtp_host          = 'your.mailserver.domain.or.ip';
$send_method        = 'sendmail'; # or 'smtp'
$admin_email_to     = "you\@yourdomain.com";
$admin_email_from   = "bhposadmin\@yourdomain.com";
$subject            = "Stock Reconcilliation";
$body_text          = '';

$VER = "1.2.2";

$PLUTYPE_UPLU=1;          # unknown and therefore slightly useless in a reconcile context.
$PLUTYPE_PLU=2;           # standard sale item
$PLUTYPE_PRICE=3;         # unknown PLU with price
$PLUTYPE_DESCPRICE=4;     # unknown plu with desc/price
$PLUTYPE_ALL=5;           # unknown plu with full entry
$PLUTYPE_RETURN=6;        # return 
$PLUTYPE_LAYBY=7;         # layby in sale
$PLUTYPE_DETAIL=8;        # detailed
$PLUTYPE_DELIVERY=9;      # delivered item
$PLUTYPE_UNDELIVERED=10;  # undelivered item
$PLUTYPE_RENTAL=11;       # rental item
$PLUTYPE_RENTALRETURN=12; # rental return item
$PLUTYPE_EPHEMERAL=13;    # ephemeral item
$PLUTYPE_PURCHASE=14;     # purchased item
$PLUTYPE_DISPATCH=15;     # order dispatch
$PLUTYPE_TRANSFERIN=16;   # transfer in
$PLUTYPE_TRANSFEROUT=17;  # transfer out

$opt_db = "bananas";
$opt_db_engine = "mysql";
$opt_host = "localhost";
$opt_user = "root";
$opt_password = "bollocks";
$opt_help = $opt_version = 0;

GetOptions("help","version","user=s","password=s",
	   "db_engine=s","db=s","host=s","new") || usage();

usage($VER) if ($opt_help || $opt_version);

$trans_processed = $items_processed = $num_sold = $num_ret = $num_rec = $num_disp = $num_tin = $num_tout =0;

$dbh = DBI->connect("DBI:$opt_db_engine:$opt_db:$opt_host",$opt_user,
		    $opt_password,{ PrintError => 0}) || die $DBI::errstr;

if($opt_new){
	generate_new_synch_entry($dbh);
   new_recstock_table($dbh);
	$dbh->disconnect;
	exit(0);
}
if(generate_synch_entry($dbh) < 0){
   if($email_notification eq 'yes' ){
      email_notice();
   }
   exit(0);
}
new_recstock_table($dbh);
generate_rectable($dbh);
update_stock($dbh);
close_synch_entry($dbh);

$body_text.="Transactions processed:\t\t$trans_processed\n";
$body_text.= "Total items:\t\t\t$items_processed\n";
$body_text.= "Sold items:\t\t\t$num_sold\n";
$body_text.= "Returned items:\t\t\t$num_ret\n";
$body_text.= "Recieved items:\t\t\t$num_rec\n"; 
$body_text.= "Dispatched items:\t\t$num_disp\n";
$body_text.= "Transfered in items:\t\t$num_tin\n";
$body_text.= "Transfered out items:\t\t$num_tout\n";
$body_text.= "\n";

if($email_notification eq 'yes' ){
   email_notice();
}

print $body_text;
$dbh->disconnect;
exit(0);

sub generate_rectable
{
	my($dbh)=@_;
	my($sth,$tth,$query);

	print "Collecting transaction data...\n";

   #generate_synch_entry($dbh);
   $ts=get_date($dbh, %ts);

	$query = "select transactionid from transaction where enteredat between '$ts' and now() order by enteredat";
	$sth = $dbh->prepare($query) or die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	while(my @r = $sth->fetchrow_array){
      process_transitems($dbh, $r[0]);
		$trans_processed++;
	}
   process_returnitems($dbh, $ts);
}

sub process_transitems
{
	my($dbh, $tid)=@_;
	my($sth,$tth,$query);

	$query = "select plu, type, quantity from transactionitems where transactionid = '$tid'";

	$sth = $dbh->prepare($query) or die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	my $count=$sth->rows;
   
   $type=0;
	while(my @row = $sth->fetchrow_array){
      $iquery = "select invid from product where plu = '$row[0]'";
	   $ith = $dbh->prepare($iquery) or die $dbh->errstr;
	   $ith->execute() or die $dbh->errstr;
      @irow = $ith->fetchrow_array;
      
      if($irow[0]){	
         $invID=$irow[0];
         $type=$row[1];
		   $query = "SELECT * FROM rec_stock WHERE plu = '$row[0]' ";

   		$tth = $dbh->prepare($query) or die $dbh->errstr;
	   	$tth->execute() or die $dbh->errstr;
	
		   my @rec_r = $tth->fetchrow_array;

   		if($rec_r[0]){	# got a row
	   		if($type==$PLUTYPE_PLU){   # sale
		   		$s = $rec_r[3] + $row[2];
			   	$tth = $dbh->prepare("UPDATE rec_stock SET sold='$s' WHERE id = '$rec_r[0]' ") || die $dbh->errstr;
				   $num_sold+=$s;
               #print "transitems sold for $tid: $count\n total:$num_sold\n";
   			}
            elsif($type==$PLUTYPE_RETURN){ # though this is now in return holding...
               # it's easier to simply increment stock in this pass, then run through
               # the return table adjusting stock in the next pass.
				   $s = $rec_r[4] + $row[2];
	   			$tth = $dbh->prepare("UPDATE rec_stock SET returned=$s WHERE id = '$r[0]' ") || die $dbh->errstr;
   				$num_ret+=$s;
               #print "transitems returned for $tid: $count\n\n total:$num_ret\n";
		   	}
			   elsif($type==$PLUTYPE_PURCHASE){
   				$s = $rec_r[5] + $row[2];
	   			$tth = $dbh->prepare("UPDATE rec_stock SET recieved=$s WHERE id = '$r[0]'") || die $dbh->errstr;
		   		$num_rec+=$s;
			   }
            elsif($type==$PLUTYPE_TRANSFERIN){
	   			$s = $rec_r[6] + $row[2];
		   		$tth = $dbh->prepare("UPDATE rec_stock SET transferin=$s WHERE id = '$r[0]'") || die $dbh->errstr;
			   	$num_rec+=$s;
	   		}
            elsif($type==$PLUTYPE_TRANSFEROUT){
		   		$s = $rec_r[7] + $row[2];
			   	$tth = $dbh->prepare("UPDATE rec_stock SET transferout=$s WHERE id = '$r[0]'") || die $dbh->errstr;
				   $num_rec+=$s;
   			}
            elsif($type==$PLUTYPE_DISPATCH){
		   		$s = $rec_r[8] + $row[2];
			   	$tth = $dbh->prepare("UPDATE rec_stock SET out=$s WHERE id = '$r[0]'") || die $dbh->errstr;
				   $num_rec+=$s;
			   }
         
   			$tth->execute || die $dbh->errstr;
	   		$items_processed++;
		   }
		   else{
            $fquery = "INSERT INTO rec_stock VALUES(0, '$invID', '$row[0]', ";
	   		if($type==$PLUTYPE_PLU){
		   		$fquery .= "'$row[2]','','','','','','')";
			   	$num_sold=$row[2];
   			}
            elsif($type==$PLUTYPE_RETURN){
		   		$fquery .= "'','$row[2]','','','',,'','')";
			   	$num_ret=$row[2];
   			}
	   		elsif($type==$PLUTYPE_PURCHASE){
		   		$fquery .= "'','','$row[2]','','','','')";
			   	$num_rec=$row[2];
   			}
            elsif($type==$PLUTYPE_TRANSFERIN){
               $fquery .= "'','','','$row[2]','','','')";
			   	$num_tin=$row[2];
			   }
            elsif($type==$PLUTYPE_TRANSFEROUT){
	   			$fquery .= "'','','','','$row[2]','','')";
		   		$num_tout=$row[2];
			   }
            elsif($type==$PLUTYPE_DISPATCH){
	   			$fquery .= "'','','','','','$row[2]','')";
		   		$num_disp=$row[2];
			   }

   			$tth = $dbh->prepare($fquery) || die $dbh->errstr;
	   		$tth->execute || die $dbh->errstr;
		   	$items_processed++;
		   }
	   }
      else{ 
         # where the fuck is the inv item?
     }
  }
}

sub process_returnitems
{
   my($dbh, $ts)=@_;
	my($sth,$tth);
  
	$sth = $dbh->prepare("SELECT * FROM return where todestat between '$ts' and NOW()") or die $dbh->errstr;
	$sth->execute || die $dbh->errstr;
	while(my @r = $sth->fetchrow_array){
      $query = "SELECT * FROM rec_stock WHERE PLU = '$row[0]'";
		$tth = $dbh->prepare($query) or die $dbh->errstr;
		$tth->execute() or die $dbh->errstr;
		my @rec_r = $tth->fetchrow_array;

		if($rec_r[0]){	# got a row
         if($r[2]=='stock'){
            $rec_r[3] = $rec_r[3] + $row[2];
            $tth = $dbh->prepare("UPDATE rec_stock SET returned=$r[3] WHERE id = $r[0]") || die $dbh->errstr;
			   $num_ret+=$r[2];
         }
         elsif($r[2]=='vendor' || $r[2]=='scrap'){
            # tits on a fish
         }
			
      }
	}
}

sub process_goodstransferitems
{
   my($dbh)=@_;
	my($sth,$tth,$rth,$pth,$num);
  
	$sth = $dbh->prepare("SELECT siteid FROM setup") or die $dbh->errstr;
	$sth->execute || die $dbh->errstr;
	my @r = $sth->fetchrow_array;
	$siteID=$r[0];
	
	$sth = $dbh->prepare("SELECT * FROM goodstransfer WHERE originator='$siteID' AND status='delivered'") or die $dbh->errstr;
	$sth->execute || die $dbh->errstr;

   # goods out transfer
	while(my @r = $sth->fetchrow_array){
	   $tth = $dbh->prepare("SELECT * FROM goodstransferitem WHERE goodstransferid = '$r[1]'") or die $dbh->errstr;
		$tth->execute() or die $dbh->errstr;
	   
	   # collect the inventory ID
	   while(my @row = $tth->fetchrow_array){
	      # find PLU in products
	      $pth = $dbh->prepare("SELECT plu FROM product WHERE invid = '$r[2]'") or die $dbh->errstr;
   		$pth->execute() or die $dbh->errstr;
	      my @prow = $pth->fetchrow_array;
	      
	      $rth = $dbh->prepare("SELECT out FROM rec_stock WHERE plu = '$prow[0]'") or die $dbh->errstr;
   		$rth->execute() or die $dbh->errstr;
	      my @rrow = $rth->fetchrow_array;
	      
	      if($rrow[0]){
	         $num=$rrow[0]+$r[3];
	         $rth = $dbh->prepare("UPDATE rec_stock SET out='$num' WHERE plu = '$prow[0]'") or die $dbh->errstr;
      		$rth->execute() or die $dbh->errstr;
	      }
	      else{
	         $rth = $dbh->prepare("INSERT INTO rec_stock VALUES(0,'$prow[0]','0','0','0','$num','0'") or die $dbh->errstr;
      		$rth->execute() or die $dbh->errstr;
	      }
	   }
	}
	
	$sth = $dbh->prepare("SELECT * FROM goodstransfer WHERE destination='$siteID' AND status='delivered'") or die $dbh->errstr;
	$sth->execute || die $dbh->errstr;

   # goods in transfer
	while(my @r = $sth->fetchrow_array){
	   $tth = $dbh->prepare("SELECT * FROM goodstransferitem WHERE goodstransferid = '$r[1]'") or die $dbh->errstr;
		$tth->execute() or die $dbh->errstr;
	   
	   # collect the inventory ID
	   while(my @row = $tth->fetchrow_array){
	      # find PLU in products
	      $pth = $dbh->prepare("SELECT plu from product where invid = '$r[2]'") or die $dbh->errstr;
   		$pth->execute() or die $dbh->errstr;
	      my @prow = $pth->fetchrow_array;
	      
	      $rth = $dbh->prepare("SELECT out from rec_stock where plu = '$prow[0]'") or die $dbh->errstr;
   		$rth->execute() or die $dbh->errstr;
	      my @rrow = $rth->fetchrow_array;
	      
	      if($rrow[0]){
	         $num=$rrow[0]+$r[3];
	         $rth = $dbh->prepare("UPDATE rec_stock SET recieved='$num' WHERE plu = '$prow[0]'") or die $dbh->errstr;
      		$rth->execute() or die $dbh->errstr;
	      }
	      else{
	         $rth = $dbh->prepare("INSERT INTO rec_stock VALUES(0,'$prow[0]','0','0','$num','0','0'") or die $dbh->errstr;
      		$rth->execute() or die $dbh->errstr;
	      }
	   }
	}
}

sub update_stock
{
	my($dbh)=@_;
	my($sth,$tth,$num);
  
	$sth = $dbh->prepare("select * from rec_stock") or die $dbh->errstr;
	$sth->execute || die $dbh->errstr;

	while(my @r = $sth->fetchrow_array){
		$tth = $dbh->prepare("SELECT instock, invid from product where plu = '$r[1]'") or die $dbh->errstr;
		$tth->execute() or die $dbh->errstr;

		my @rec_row = $tth->fetchrow_array;
		$num=$rec_row[0]+$r[2]-$r[3]+$r[4]-$r[5];

		$tth = $dbh->prepare("UPDATE product SET instock='$num' where plu = '$r[1]'") || die $dbh->errstr;
		$tth->execute || die $dbh->errstr;
		
		$tth = $dbh->prepare("UPDATE inventory SET instock='$num' where id = '$rec_row[1]'") || die $dbh->errstr;
		$tth->execute || die $dbh->errstr;
		
		$tth = $dbh->prepare("SELECT rolevel, onorder from inventory where id = '$rec_row[1] AND \
         (instock + onorder) < rolevel'") or die $dbh->errstr;
		$tth->execute() or die $dbh->errstr;

		my @wank_row = $tth->fetchrow_array;
		if($wank_row[0] < ($wank_row[1] + $num)){
		   $dis=($wank_row[1] + $num) - $wank_row[0];
		   $tth = $dbh->prepare("UPDATE rec_stock SET stocklevelwarning='$dis' WHERE plu = '$r[1]'") || die $dbh->errstr;
   		$tth->execute || die $dbh->errstr;
		}
	}
}

sub get_date
{
	my($dbh,$ts)=@_;
	my($sth,$query);

	$sth = $dbh->prepare("SELECT  implementationat from synch where implemented='Y' and typeid='3' ORDER BY  implementationat DESC LIMIT 1") || die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	if($sth->rows==0){
		print "The synch table has no last recstock entry - quitting\n";
		exit(0);
	}
	my @r = $sth->fetchrow_array;
	$ts=$r[0];
	
	#print "$ts\n";
	
	return $ts;
}

sub new_recstock_table
{
	my($dbh)=@_;
	my($sth,$query);
	
	$sth = $dbh->prepare("select count(*) from rec_stock") or die $dbh->errstr;
	if($sth->execute){
		$sth = $dbh->prepare("DELETE FROM rec_stock") or die $dbh->errstr;
		$sth->execute() or die $dbh->errstr;
	}
	else{
		$query = "CREATE TABLE rec_stock (";
	   $query .= "id int(11) not null auto_increment,";
      $query .= "invid int(11) not null,";
	   $query .= "plu varchar(16) not null,";
	   $query .= "sold int(11) not null,";
	   $query .= "returned int(11) not null,";
	   $query .= "recieved int(11) default '0' not null,";
      $query .= "transferin int(11) default '0' not null,";
      $query .= "transferout int(11) default '0' not null,";
	   $query .= "out int(11) default '0' not null,";
	   $query .= "stocklevelwarning int(11) NOT NULL,";
 	   $query .= "PRIMARY KEY (id),";
 	   $query .= "UNIQUE (plu))";

		$sth = $dbh->prepare($query) or die $dbh->errstr;
		$sth->execute() or die $dbh->errstr;
      
      generate_new_synch_entry();
	}
}

# put the synch entry in on start in case of interrupts
sub generate_synch_entry
{
	my($dbh)=@_;
	my($sth);

	$sth = $dbh->prepare("SELECT * FROM synch WHERE implemented='N' AND typeid='3'") || die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	if($sth->rows){
		print "The synch table has outstanding entries, cannot perform operation\n";
		return -1;
	}

	$sth = $dbh->prepare("INSERT INTO synch VALUES(0,'3',NOW(),NOW(),'','N')") || die $dbh->errstr;
	$sth->execute || die $dbh->errstr;
   return 0;
}

# close synch entry
sub close_synch_entry
{
	my($dbh)=@_;
	my($sth, $tth);

	$sth = $dbh->prepare("SELECT ID FROM synch WHERE implemented='N' AND typeid='3'") || die $dbh->errstr;
	$sth->execute() or die $dbh->errstr;
	if($sth->rows){
      my @r = $sth->fetchrow_array;
		$tth = $dbh->prepare("UPDATE synch SET implemented='Y', implementationat=NOW() WHERE id= $r[0]") || die $dbh->errstr;
	   $tth->execute || die $dbh->errstr;
		exit(0);
	}
}

# new option
sub generate_new_synch_entry
{
	my($dbh)=@_;
	my($sth);

	$sth = $dbh->prepare("INSERT INTO synch VALUES(0,'3',NOW(),NOW(),NOW(),'Y')") || die $dbh->errstr;
	$sth->execute || die $dbh->errstr;
}

sub email_notice
{
   if($send_method eq 'smtp'){
      # use this for windows machines that don't have sendmail
      MIME::Lite->send("$send_method", "$mailprog_or_smtp_host", Timeout=>60);
      $msg = new MIME::Lite
             From     =>"$admin_email_from",
             To       =>"$admin_email_to",
             Subject  =>"$subject",
             Type     =>"TEXT",
             Encoding =>"7bit",
             Data     =>"$body_text"; 
  
      $msg->send || die print qq~Error!\n\nError in Mailing Program!~;
   }
   else{
      &mail_to($admin_email_to, $admin_email_from, $subject, $body_text, $admin_email_from);
   }
}

sub mail_to
{
   # &mail_to($email_to, $email_from, $subject, $mail_body, $reply_to);
   my($email_to, $email_from, $subject, $mail_body, $reply_to) = @_;

   print "Sending email notification\n";
   
   if($reply_to !~ /\@/ ){$reply_to = $email_from;}

   open(MAIL, "|$mailprog") || die print qq~Error!\n\nCan't open $mailprog!~;

   print MAIL "To: $email_to\n";
   print MAIL "From: $email_from\n";
   print MAIL "Subject: $subject\n";
   print MAIL "Reply-To: $reply_to\n";
   print MAIL "\n";
   print MAIL "$mail_body";
   print MAIL "\n";
   close (MAIL);
}

sub usage
{  
  my($VER)=@_;
  
  $0 =~ s/.\/(.+)/$1/;
  if ($opt_version)
  {
    print "$0 version $VER\n";
  } 
  else
  {
    print <<EOF;
$0 version $VER

Usage: $0 [options] 

Description: Reconcile stock

Options:
--help              show this help and exit
--version           shows the version of the program
--db_engine=...     database server (default: $opt_db_engine)
--db=...            database to be used (default: $opt_db)
--force=...         force a reconcile (not recommended as synch will be lost)     
--host=...          hostname to be used (default: $opt_host)
--new=...           generate an initial entry
--password=...      user password for the db server
--user=...          username for the db server
EOF
  }
  exit(0);
}
